# Config Actions

## What is a config action?

A config action is a way of altering configuration. The main initial use case
of config actions is to alter existing configuration in a Drupal [recipe](recipe.md).

Say a module or another recipe has provided a piece of configuration. Now a
recipe you're applying needs to alter that configuration. That's what a config
action is for.

The classic use case is to add one or more permissions to a user role. Say
you're applying an `article` recipe that provides an `article` content type.
Before it runs, the `article` recipe applies a number of other recipes, one of
which provides an `editor` role. Now the `article` recipe needs to add article-
related permissions to that existing `editor` role. Enter config actions.

The configuration actions API proposed for Drupal core provides a flexible
system for creating different types of actions. Initially, there are two sets
of actions supported by the API.

1. __Actions that can be applied to configuration of any configuration entity
   type.__
2. __Actions declared for methods of specific configuration entity types.__
   These actions are created by adding [PHP attributes](https://www.php.net/manual/en/language.attributes.php) to a method declaration.

Config entity actions can have a plural as well as a singular form. For
example, the plural form of the `grantPermission` action is `grantPermissions`.

### Format of config actions

Config actions are declared in the `recipe.yml` file of a Drupal recipe.

The structure for each action has three parts:

1. _Configuration entity ID_.
2. _The action_, which must be one that's declared for the entity type of the
   configuration entity.
3. _Action arguments_, which take different forms depending on:
    * the number of arguments that the action takes, and
    * whether the action is in the singular or plural form.

The basic format for a singular action that takes only one argument is:

```yaml
config:
  actions:
    some.config.id:
      someAction: 'some argument'
```

The `grantPermission` action is derived from the
`\Drupal\user\Entity\Role::grantPermission()` method which, per [its API documentation](https://api.drupal.org/api/drupal/core%21modules%21user%21src%21Entity%21Role.php/function/Role%3A%3AgrantPermission/10.0.x),
takes only a single argument, `$permission`. So here's how the action looks
in the singular when used to grant a permission to the `editor` user role:

```yaml
config:
  actions:
    user.role.editor:
      grantPermission: 'delete any article content'
```

In the plural form, the argument is not a single string but an array of them,
each of which will be applied to a separate `::grantPermission()` call. Here's
how that looks:

```yaml
config:
  actions:
    user.role.editor:
      grantPermissions:
        - 'delete any article content'
        - 'edit any article content'
```

When the action's method accepts more than one argument, the arguments are
passed as an array. An example is the `setComponent` action for entity view and form displays. 
The [`EntityDisplayBase::setComponent()` method](https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Entity%21EntityDisplayBase.php/function/EntityDisplayBase%3A%3AsetComponent/10.0.x) takes two arguments:

> string `$name`: The name of the component.
> array `$options`: The display options.

So in the singular the action passes a list of arguments, first a string and
then an array. A sample action adding `field_tags` configuration to the
`default` view display for the `article` content type might look like this:

```yaml
config:
  actions:
    core.entity_view_display.node.article.default:
      setComponent:
        name: field_tags
        options:
          type: entity_reference_label
          label: above
          settings:
            link: true
          third_party_settings: {  }
          weight: 10
          region: content
```

A pluralized version, adding the configuration for multiple fields to the same
display, would pass an array of arrays:

```yaml
config:
  actions:
    core.entity_view_display.node.article.default:
      setComponents:
        -
          name: field_tags
          options:
            type: entity_reference_label
            label: above
            settings:
              link: true
            third_party_settings: {  }
            weight: 10
            region: content
        -
          name: field_categories
          options:
            type: entity_reference_label
            label: above
            settings:
              link: true
            third_party_settings: {  }
            weight: 11
            region: content
```

## Adding support for new config entity actions

To create an action for an existing config entity method, you add a PHP
attribute to the method declaration. At a minimum, you need to provide an
admin label for the action.

For example, the `Role::grantPermission()` method is exposed as a config
action through the following code that comes immediately before the function
declaration (`public function grantPermission($permission) {`):

```php
#[ActionMethod(adminLabel: new TranslatableMarkup('Add permission to role'))]
```

The PHP attribute uses two class names. Depending on whether the class file
you're editing already has them, you may need to add a `use` statement for one
or both of these classes. Example:

```php
use Drupal\Core\Config\Action\Attribute\ActionMethod;
use Drupal\Core\StringTranslation\TranslatableMarkup;
```

By default, a pluralized version of the action name will also be generated
using [Symfony's English string inflector](https://symfony.com/doc/current/components/string.html#inflector).

If the method name doesn't lend itself to automatic pluralization, you can
explicitly pass a plural form. Take for example a `::set()` method. Since
`sets` doesn't fit here, the PHP attribute could be:

```php
#[ActionMethod(adminLabel: new TranslatableMarkup('Set a property on the configuration item'), pluralize: 'setMultiple')]
```

If it doesn't make logical sense to support a pluralized form, you can pass a
value of `FALSE`:

```php
#[ActionMethod(adminLabel: new TranslatableMarkup('Do something'), pluralize: FALSE)]
```

For notes on how to determine what config actions are needed for particular
config entity changes, see the section "Convert config overrides to config
actions" in the [recipe author guide](recipe_author_guide.md).

## Available config actions

### Applicable to all config entity types

#### Simple config update

A globally applicable config action, `simple_config_update`, can be used to
make updates to any configuration value. Despite its name, this action can be
applied to any type of configuration. (There's a `@todo` in the code
questioning whether this is correct.)

Sample usage to update the `page.front` value in the `system.site` simple
configuration:

```yaml
actions:
  system.site:
    simple_config_update:
      page.front: '/node'
```

The plural version takes the same name as the singular but passes multiple
arguments. Example:

```yaml
actions:
  system.site:
    simple_config_update:
      page.front: '/node'
      default_langcode: 'es'
```

#### Ensure exists

The `ensure_exists` config action can be used to conditionally create a config
entity. However, it's not clear when this is really a recommended approach; see
[Clarify recommended use case(s) for ensure_exists config action](https://www.drupal.org/project/distributions_recipes/issues/3306291).

Pending a confirmed use case, it's not generally recommended to use
`ensure_exists`. Instead, if you want to ensure configuration such as a role
exists without directly providing it, apply another recipe that provides the
configuration.

#### Create

The `create` config action can be used to create a config entity.
Like `ensure_exists`, `create` is not likely to be often useful in recipes.

#### Configure Third Party Settings
The `setThirdPartySetting(s)` config actions allows the recipe author to set third-party 
settings values for any config entity that can carry third-party settings, which at the time 
of this writing includes all config entities. Helpful for modules like Layout Builder and Scheduler 
that use these settings to provide or change functionality.

Singular:

```yaml
config:
  actions:
    core.entity_view_display.node.test.full:
      setThirdPartySetting:
        module: layout_builder
        key: enabled
        value: true

```

To set multiple third-party settings:

```yaml
config:
  actions:
    core.entity_view_display.node.test.full:
      setThirdPartySettings:
        -
          module: layout_builder
          key: enabled
          value: true
        -
          module: layout_builder
          key: allow_custom
          value: true
```

### Applicable to specific config entity types

These config actions can be applied to specific config entity types. Further
config actions are being added as sub-issues of [Determine which core config
entity methods should be config actions](https://www.drupal.org/project/distributions_recipes/issues/3303127).

#### Config Entity Wildcards

You can use wildcards in your recipe to apply actions to multiple config entities of the same type.

Here are some examples of possible uses:

##### Adding the same permission(s) to all user roles

```yaml
actions:
  user.role.*:
    grantPermissions:
      - 'access content'
```

##### Affecting every view display of every content type

```yaml
actions:
  core.entity_view_display.node.*.*:
    setComponent:
      name: field_tags
      options:
        type: entity_reference_label
        label: above
        settings:
          link: true
        third_party_settings: {  }
        weight: 10
        region: content
```

##### Changing settings on every instance of a field

```yaml
actions:
  field.field.media.*.field_tags:
    simple_config_update:
      foo: bar
```

##### Creating a view display (i.e., enabling a view mode) for every content type

```yaml
actions:
  core.entity_view_display.node.*.foo:
    ensure_exists:
      # The contents of the view display here...
```

##### Overriding a base field for every media type

```yaml
actions:
  core.base_field_override.media.*.thumbnail:
    ensure_exists:
      # The contents of the base field override here...
```

Note that you cannot use wildcards on simple config (such as `gin.settings`), or all config provided by a specific module. For example, something like `node.*` to target _all_ config provided by the Node module, won't work.

You also cannot use wildcards on config prefixes. For example, these are all invalid and will raise exceptions:

```yaml
# Won't hit field.field and field.storage.
field.*.media.foo.field_bar
# Won't hit all configs provided by a project.
node.*.something
# Won't hit config from any module that provides something called a "vocabulary".
*.vocabulary.wat
```

In other words, you can only put wildcards in the "identifying" parts of a config entity's full name. 
This is due to the way config entities relate to the underlying config schema.


#### User role permissions

* _Config entity type_: [`user_role`](https://api.drupal.org/api/drupal/core%21modules%21user%21src%21Entity%21Role.php/class/Role/10.0.x)
* _Action name_:
    * _singular_: `grantPermission`
    * _plural_: `grantPermissions`

User roles also have specialized actions that allow you to grant permissions
that are specific to a particular bundle -- for example, permissions that are
defined per content type, or per media type, such as `create page content` or
`edit own image media`. If you want to grant one of these permissions for
_every_ bundle of an entity type, you can use config actions to do it dynamically.

For example, to grant the ability for a role to edit their own content, for every
content type:

```yaml
config:
  actions:
    user.role.editor:
      grantPermissionsForEachNodeType:
        - 'edit own %bundle content'
```

The token `%bundle` is dynamically replaced by the machine name of the bundle.

It also works for media types:

```yaml
config:
  actions:
    user.role.media_creator:
      grantPermissionsForEachMediaType:
        - 'create %bundle media'
        - 'delete own %bundle media'
```

And taxonomy vocabularies:

```yaml
config:
  actions:
    user.role.info_architect:
      grantPermissionsForEachTaxonomyVocabulary:
        - 'create %bundle terms'
```

**Advanced**: Any entity type that supports bundles will automatically have
actions like this. The name of the action is derived from the machine name
(ID) of the bundle entity type. So say that `paragraph` entities' bundles
are defined by `paragraph_type` entities; this will result in a config action
called `grantPermissionsForEachParagraphType`, which allows you to grant
paragraph type-specific permissions to a user role.

#### Adding components to entity view and form displays

As mentioned above, to add a field to an entity's view or form displays, you can use 
`setComponent(s)`.

Singular:

```yaml
config:
  actions:
    core.entity_view_display.node.article.default:
      setComponent:
        name: field_tags
        options:
          type: entity_reference_label
          label: above
          settings:
            link: true
          third_party_settings: {  }
          weight: 10
          region: content
```

A pluralized version, adding the configuration for multiple fields:

```yaml
config:
  actions:
    core.entity_view_display.node.article.default:
      setComponents:
        -
          name: field_tags
          options:
            type: entity_reference_label
            label: above
            settings:
              link: true
            third_party_settings: {  }
            weight: 10
            region: content
        -
          name: field_categories
          options:
            type: entity_reference_label
            label: above
            settings:
              link: true
            third_party_settings: {  }
            weight: 11
            region: content
```

#### Adding a field to every bundle of an entity type

If you have a field that you want to add to every content type, but you don't know
how many content types there are, or what they're called, there's an `addToAllBundles`
config action that will automatically add a field to every content type.

For example, if you have a field storage called `field.storage.node.field_meta_tags`, and
you want to add it to every content type:

```yaml
config:
  actions:
    field.storage.node.field_meta_tags:
      addToAllBundles:
        label: Meta tags
        description: Choose your meta tags for this content.
```

This action works exactly the same way for fields that attach to media types, taxonomy terms,
or any other fieldable entity type. You can specify the field's label and user-facing
description, and it will be the same for every bundle. At the time of this writing, you
_cannot_ change the settings of the new fields; they will be created with their default
settings.

By default, the `addToAllBundles` action will not conflict with fields that already exist.
So, for example, if you have `field_meta_tags` tags on your `article` content type, but not
your `page` and `blog_post` content types, the action will add the field to the `page` and
`blog_post` content types, but the one on `article` will be left alone. However, if you
specify the `fail_if_exists` flag, an error will occur if the field already exists on
any content type:

```yaml
config:
  actions:
    field.storage.node.field_meta_tags:
      addToAllBundles:
        label: Meta tags
        description: Choose your meta tags for this content.
        # If field_meta_tags already exists on any content type, you'll get an error.
        fail_if_exists: true
```

Another thing to bear in mind is that this action will add the field to every bundle, but
it will NOT make the field visible to editors or users. To do that, you will also need to
use the `setComponent` action of entity view displays and entity form displays. Here, then,
is a complete example of adding a field to every content type, and then ensuring it's available
to editors and viewers:

```yaml
config:
  actions:
    field.storage.node.field_byline:
      addToAllBundles:
        label: Byline
        description: The name of the person who wrote this content.
    # Make sure the field is visible on the node edit forms.
    entity_form_display.node.*.default:
      setComponent:
        field_byline:
          type: text_textfield
    # Ensure the field is visible when viewing content.
    entity_view_display.node.*.full:
      setComponent:
        field_byline:
          type: text_default
```

#### Adding entity types and bundles to a Content Moderation workflow

If you have an entity type or entity bundle that you want to add to a content moderation
workflow, you can use config actions.

```yaml
config:
  actions:
    workflows.workflow.editorial:
      # Opt only certain content types into moderation.
      addNodeTypes: [page, article]
      # Opt every taxonomy vocabulary into moderation.
      addTaxonomyVocabularies: '*'
```

The action is dynamic and automatically converts singular entity type plugin IDs, like 
`taxonomy_vocabulary`, into plural strings like `TaxonomyVocabularies` so it will work with 
custom and additional config entity types. Just "pluralize" the plugin ID, remove any underscores, convert to SentenceCase, and prepend `add`. For example, the `paragraph_type`
entity type will become `addParagraphTypes`.

#### Add a button to the toolbar of a CKEditor 5 editor

To add a new toolbar item, optionally at a specific position, and optionally 
replacing the item that's already in that position, use the `addItemToToolbar` config action.

```yaml
config:
  actions:
    editor.editor.foo:
      addItemToToolbar:
        item_name: MediaLibrary
        # Specific zero-based position in the toolbar.
        # If not specified, it appends to the end.
        position: 3
        # If there's already something at that position, replace it.
        # For example, replace the Image button with Media library. By default,
        # existing items are left where they are, and the new item is inserted
        # right before them.
        replace: true
```
